// (c) Peter Kankowski, 2007. http://smallcode.weblogs.us mailto:kankowski@narod.ru
#include <assert.h>
#include <stdio.h>
#include <windows.h>

// ================================
//   Code 
// ================================
enum OPERATION_CODE {
	OP_ADD = 0xC1DE, // FADDP instruction
	OP_SUB = 0xE9DE,
	OP_MUL = 0xC9DE,
	OP_DIV = 0xF9DE
};

class ExprCode {
private:
	short*  _code;
	short*  _cur_code;
	short*  _code_end;

	double* _data;
	double* _cur_data;
	double* _data_end;

	void PrintErrorAndExit() {
		puts("Out of memory");
		exit(1);
	}
	static const short OP_FLD  = 0x05DDu;
	static const short OP_FSTP = 0x1DDDu;
	static const short OP_RET = 0x90C3u; // Emit RET and NOP
public:
	ExprCode() {
		SYSTEM_INFO si;
		GetSystemInfo(&si);
		DWORD region_size = si.dwAllocationGranularity;
		//region_size = UINT_MAX; // To test the error-handling code
		// Allocate two regions: one for code, and one for data
		_code = (short*)VirtualAlloc(NULL, region_size * 2, MEM_COMMIT, PAGE_EXECUTE_READWRITE);
		if(!_code)
			PrintErrorAndExit();
		_code_end = (short*) ((char*)_code + region_size);
		_data = (double*)_code_end;
		_data_end = (double*)((char*)_data + region_size);
		// Initialize current pointers
		_cur_code = _code;
		_cur_data = _data;
	}
	~ExprCode() {
		VirtualFree(_code, NULL, MEM_RELEASE);
	}
	void EmitOperation(OPERATION_CODE opcode) {
		if(_cur_code >= _code_end)
			PrintErrorAndExit();
		*_cur_code = (short)opcode;
		_cur_code++;
	}
private:
	void EmitFldFstp(short opcode, double* ptr) {
		if((char*)_cur_code + sizeof(int) >= (char*)_code_end)
			PrintErrorAndExit();
		*_cur_code++ = opcode;
#if defined(_WIN64)
		// Under x86-64, the address is relative to the address of the next instruction
		*(int*)_cur_code = (int)(INT_PTR)((char*)ptr - ((char*)_cur_code + sizeof(int)));
#elif defined(_WIN32)
		// Under x86, flat 32-bit address is used
		*(int*)_cur_code = (int)(INT_PTR)ptr;
#else
#error Unsupported platform
#endif
		_cur_code = (short*)((char*)_cur_code + sizeof(int));
	}
public:
	void EmitLoadNumber(double number) {
		if(_cur_data >= _data_end)
			PrintErrorAndExit();
		// Write the number to data section
		*_cur_data = number;
		// Emit FLD [_cur_data] instruction and advance the pointer
		EmitFldFstp(OP_FLD, _cur_data);
		_cur_data++;
	}
	double* EmitStoreNumber() {
		if(_cur_data >= _data_end)
			PrintErrorAndExit();
		// Emit FSTP [_cur_data]
		double* num = _cur_data;
		EmitFldFstp(OP_FSTP, num);
		// Advance the pointer
		_cur_data++;
		return num;
	}
	void EmitEnd() {
		if(_cur_code >= _code_end)
			PrintErrorAndExit();
		*_cur_code = OP_RET;
		_cur_code++;
	}
	void Execute() {
		// Cast _code to the pointer to a function returning void
		void (*p)() = (void (*)())_code;
		p();
	}
};

#ifdef _DEBUG
void TestExprCode() {
	ExprCode code;
	// -3.5 + 1.2 = -2.3
	code.EmitLoadNumber(-3.5);
	code.EmitLoadNumber(1.2);
	code.EmitOperation(OP_ADD);
	double *res1 = code.EmitStoreNumber();

	// (1.5 - 0.5 + 0.2) / 5 = 0.24
	code.EmitLoadNumber(1.5);
	code.EmitLoadNumber(0.5);
	code.EmitOperation(OP_SUB);
	code.EmitLoadNumber(0.2);
	code.EmitOperation(OP_ADD);
	code.EmitLoadNumber(5);
	code.EmitOperation(OP_DIV);
	double *res2 = code.EmitStoreNumber();

	// (1 - 10^6 / 10^5) * 2.5 = -22.5
	code.EmitLoadNumber(1);
	code.EmitLoadNumber(10e6);
	code.EmitLoadNumber(10e5);
	code.EmitOperation(OP_DIV);
	code.EmitOperation(OP_SUB);
	code.EmitLoadNumber(2.5);
	code.EmitOperation(OP_MUL);
	double *res3 = code.EmitStoreNumber();
	
	// Load and store one number
	code.EmitLoadNumber(-1.23e5);
	double *res4 = code.EmitStoreNumber();

	code.EmitEnd();
	code.Execute();
	assert(*res1 == -2.3);
	assert(*res2 == 0.24);
	assert(*res3 == -22.5);
	assert(*res4 == -1.23e5);
}
#endif

int main() {
#ifdef _DEBUG
	TestExprCode();
	puts("The program passed all tests.");
#endif
}